import os
from typing import List

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.io import arff
from sklearn.metrics import roc_curve, auc
from sklearn.preprocessing import OrdinalEncoder

show_plots = False

# Plots

def heat_map(df, model_name=None) -> None:
    """
    Generates a heatmap of the correlation matrix for numerical features in the DataFrame.
    :param df: The given dataframe.
    :param model_name: The name of the model to use.
    """
    # Drop all NaN
    df.dropna(axis='columns')
    ndf = df[[col for col in df.columns if df[col].nunique() > 1 and pd.api.types.is_numeric_dtype(df[col])]]
    corr = ndf.corr()
    plt.figure(figsize=(15, 12))
    sns.heatmap(corr)
    if model_name:
        save_plot(model_name + " - Data Correlations")
    if show_plots:
        plt.show()

def plot_xy(df, x, y, model_name=None) -> None:
    """
    Creates a scatter plot for two numerical columns.
    :param df: The given dataframe.
    :param x: First feature name to be plotted in the x-axis.
    :param y: Second feature name to be plotted in the y-axis.
    :param model_name: The name of the model to use.
    """
    plt.scatter(df[x], df[y])
    plt.xlabel(x)
    plt.ylabel(y)
    if model_name:
        save_plot(model_name + " - XY")
    if show_plots:
        plt.show()

def plot_features(features, info_text: str = None, model_name=None) -> None:
    """
    Displays a bar graph with the importance of each feature.
    :param features: The given dataframe with all feature importance.
    :param info_text: Optional text in the legend.
    :param model_name: The name of the model to use.
    """
    plt.figure(figsize=(10, 10))

    # Add labels to bars
    for index, value in enumerate(features['Importance']):
        plt.text(value + 0.005, index, f'{value:.4f}', fontsize=10, verticalalignment='center')

    plt.barh(features.index, features.Importance)
    plt.xlabel('Importance')
    plt.title('Feature Importance')
    plt.gca().invert_yaxis()

    # Fix text cutoff issue
    plt.subplots_adjust(left=0.3)  # ⬅ Increase left margin for feature names
    plt.tight_layout()

    if info_text:
        x_max = plt.xlim()[1]
        y_max = plt.ylim()[0]
        plt.text(x_max, y_max, info_text, verticalalignment='bottom', horizontalalignment='right')
    if model_name:
        save_plot(model_name + " - Feature Importance")
    if show_plots:
        plt.show()

def plot_confusion_matrix(confusion_matrix: List[List[int]], accuracy: float, model_name=None) -> None:
    """
    Plots the confusion matrix as a heatmap.
    :param confusion_matrix: The given confusion matrix.
    :param accuracy: Accuracy score of the given confusion matrix.
    :param model_name: The name of the model to use.
    :return:
    """
    if len(confusion_matrix) != 2 or any(len(row) != 2 for row in confusion_matrix):
        raise ValueError("Confusion matrices must be 2x2")

    plt.figure(figsize=(8, 6))
    labels = [[f'True Positive\n{confusion_matrix[0][0]}', f'False Positive\n{confusion_matrix[0][1]}'],
              [f'False Negative\n{confusion_matrix[1][0]}', f'True Negative\n{confusion_matrix[1][1]}']]
    plt.title(f"Confusion matrix | Accuracy: {accuracy * 100:.2f}%")
    sns.heatmap(confusion_matrix, annot=labels, fmt='', cmap='Blues', cbar=False)
    if model_name:
        save_plot(model_name + " - Confusion Matrix")
    if show_plots:
        plt.show()

def plot_precision_recall_curve(precision, recall, model_name=None):
    """
    A Precision-Recall curve shows the trade-off between precision
    (how many predicted positives are actually correct) and recall
    (how many actual positives were correctly identified).
    :param precision: Precision of the model.
    :param recall: Recall of the model.
    :param model_name: The name of the model to use.
    :return:
    """
    plt.figure(figsize=(8, 6))
    plt.plot(recall, precision, marker='.', label="Precision-Recall Curve")
    plt.xlabel("Recall")
    plt.ylabel("Precision")
    plt.title("Precision-Recall Curve")
    plt.legend()
    plt.grid()
    if model_name:
        save_plot(model_name + " - Precision Recall Curve")
    if show_plots:
        plt.show()

def plot_learning_curve(train_sizes, train_scores, test_scores, model_name=None):
    """
    A learning curve helps diagnose overfitting or underfitting by plotting
    training and validation performance as training size increases.
    :param train_sizes: The train sizes of the model.
    :param train_scores: The train scores of the model.
    :param test_scores: The test scores of the model.
    :param model_name: The name of the model to use.
    """
    train_mean = np.mean(train_scores, axis=1)
    train_std = np.std(train_scores, axis=1)
    test_mean = np.mean(test_scores, axis=1)
    test_std = np.std(test_scores, axis=1)

    plt.figure(figsize=(8, 6))
    plt.plot(train_sizes, train_mean, 'o-', label="Training Score", color="blue")
    plt.plot(train_sizes, test_mean, 'o-', label="Validation Score", color="red")

    plt.fill_between(train_sizes, train_mean - train_std, train_mean + train_std, alpha=0.1, color="blue")
    plt.fill_between(train_sizes, test_mean - test_std, test_mean + test_std, alpha=0.1, color="red")

    plt.xlabel("Training Size")
    plt.ylabel("Accuracy")
    plt.title("Learning Curve")
    plt.legend()
    plt.grid()
    if model_name:
        save_plot(model_name + " - Learning Curve")
    if show_plots:
        plt.show()

def plot_counts(target, df) -> None:
    """
    Display a bar graph counting the quantity of the target variable.
    :param target: The target variable in the dataframe.
    :param df: The given dataframe.
    """
    plt.clf()
    sns.countplot(x = target, data = df)
    save_plot("Count")
    if show_plots:
        plt.show()

def plot_roc_curve(y_true, y_score, model_name=None) -> None:
    """
    Plots the ROC curve for a binary classification model.
    :param y_true: The true labels.
    :param y_score: The predicted probabilities.
    :param model_name: The name of the model to use.
    :return:
    """
    fpr, tpr, _ = roc_curve(y_true, y_score)
    roc_auc = auc(fpr, tpr)

    plt.figure(figsize=(8, 6))
    plt.plot(fpr, tpr, color='darkorange', lw=2, label=f'ROC curve (area = {roc_auc:.2f})')
    plt.plot([0, 1], [0, 1],color='navy', lw=2, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic (ROC) Curve')
    plt.legend(loc="lower right")
    plt.grid()
    if model_name:
        save_plot(model_name + " - ROC Curve")
    if show_plots:
        plt.show()

def save_plot(name) -> None:
    """
    Saves the current plot.
    :param name: The name of the model to use.
    """
    os.makedirs("resulting_figures", exist_ok=True)
    plt.savefig("resulting_figures/" + name, dpi=300, bbox_inches='tight')

# Data processing

def import_data(train_file_path, test_file_path, model_name):
    """
    Imports data from the given path.
    :param train_file_path: The path to the training data file.
    :param test_file_path: The path to the test data file.
    :param model_name: The name of the model to use.
    """
    data, meta = arff.loadarff(train_file_path)
    df_train = pd.DataFrame(data)
    df_train = df_train.map(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)

    data, meta = arff.loadarff(test_file_path)
    df_test = pd.DataFrame(data)
    df_test = df_test.map(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)

    return df_train, df_test, model_name

def ordinal_encode(df, categories, target) -> None:
    """
    Applies ordinal encoding to a specified categorical column in a DataFrame.
    :param df: The given dataframe.
    :param categories: The possible values of target.
    :param target: The column to be encoded.
    :raises TypeError: Occurs when there are more than two possible values for target.
    """
    if categories is None or len(categories) != 2:
        raise TypeError("Categories must be provided")
    df[target] = OrdinalEncoder(categories = [categories]).fit_transform(df[[target]])

def normalize(df_train, df_test, exclude, numerical_scaler, label_scaler) -> None:
    """
    Normalize the data in the given dataframe.
    :param df_train: The training dataframe.
    :param df_test: The test dataframe.
    :param exclude: Columns to be excluded from normalisation.
    :param numerical_scaler: A scaler to normalize numerical values.
    :param label_scaler: A scaler to normalize categorical values.
    :return:
    """
    df_temp = pd.concat([df_train, df_test])
    scale_targets = df_temp.select_dtypes(include=np.number).drop(columns=exclude).columns
    numerical_scaler.fit_transform(df_temp[scale_targets])

    df_train[scale_targets] = numerical_scaler.transform(df_train[scale_targets])
    df_test[scale_targets] = numerical_scaler.transform(df_test[scale_targets])

    labels = df_train.select_dtypes(include=object, exclude=np.number).columns
    for label in labels:
        label_scaler.fit_transform(df_temp[label])
        df_train[label] = label_scaler.transform(df_train[label])
        df_test[label] = label_scaler.transform(df_test[label])

# Additional metrics

def print_high_confidence_samples(model, x: pd.DataFrame) -> None:
    """
    Prints in the output stream samples for which the model is 90% confident about.
    :param model: The given model.
    :param x: The features of the dataframe.
    """
    # Get predicted probabilities
    predicted_probabilities = pd.DataFrame(model.predict_proba(x)[:, 1],
                                           columns=['confidence level'])  # Probability of being class 1
    # Filter samples where the model is at least 90% sure
    high_confidence_samples = predicted_probabilities[predicted_probabilities['confidence level'] > 0.9]
    print(high_confidence_samples.head())